from RciPacket import RciPacket
from RciLog import RciLog
import socket
from io import BufferedReader,BufferedWriter
from array import array
from posix import strerror
import os
from socket import errno
import codecs
from _socket import timeout
from _struct import unpack, unpack_from

class RciClient():
    
    SOCKET_CONNECTION_TIMEOUT_SEC = 5 
    SOCKET_RECV_BUFFER_SIZE = 4096
    log = RciLog("RCI Client")
    socketConnected = False

    def __init__(self):
        try :
            self.sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            self.log.write("Create Client Socket",'NOTIFICATION')
        except socket.error as e:
            self.log.write("Unable to create socket : " + os.strerror(e.errno) ,'ERROR')

        
    def getPort(self):
        return self.sock.getsockname()[1]
    
    def getHostName(self):
        return self.sock.gethostbyname(self.sock.gethostname())()

    def setLog(self,l):
        self.log = l

    # Setup connection to server
    def connect(self, host, port):
        try:
            self.port = port
            self.host = host
            if self.sock != None:
                if self.socketConnected==False:
                    self.sock.settimeout(self.SOCKET_CONNECTION_TIMEOUT_SEC)
                    self.sock.connect((host,port))
                    self.socketConnected=True
                else:
                    self.log.write("Connection already established",'WARNING')
            else :
                self.log.write("No Socket Object",'ERROR')
                raise Exception
            self.log.write("Connection to server " + str(host) + " established on port " + str(port), 'NOTIFICATION')
        except socket.error as s:
            self.log.write("connect :  socket error : " + os.strerror(s.errno) ,'ERROR')
            raise s
        except Exception:
            self.log.write("connect : exception ",'ERROR')
            raise Exception
        except timeout:
            self.log.write("connect : timeout ",'ERROR')
            raise timeout("Socket Timeout")

    def isConnected(self):
        if self.sock != None and self.socketConnected:
            return True
        return False
        
    def disConnect(self):
        if self.sock != None:
            if self.socketConnected==True:
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
                self.socketConnected=False
                self.log.write("Disconnection from server",'NOTIFICATION')
            else:
                self.log.write("Connection already disconnected",'WARNING')
        else:
            self.log.write("No Socket Object",'ERROR')
            raise Exception

    # Sent a packet request to the server
    def sendRequest(self, packetRequest):
        if (self.isConnected() == False):
            self.log.write("Connection to " + self.host + " is lost or not yet established",'ERROR')
            raise Exception

        try :
            #packetRequest.clearPacket()
            tosendBuffer = array('c','\0'*RciPacket.RCI_MIN_NUM_BYTES_PACKET)
            packetRequest.setCurrentTimeStamp()
            packetRequest.packetWrite(tosendBuffer)
            if self.sock==None:
                self.log.write("No Socket Object",'ERROR')
                raise Exception
            self.sock.sendall(tosendBuffer.tostring())
            self.log.write("Request sent", 'NOTIFICATION')
        except socket.error as s:
            self.log.write("Unable to send request : " + os.strerror(s.errno) ,'ERROR')
            raise s
        except Exception as e:
            raise e

    # Receive a packet reply from the server
    def ReceiveReply(self, packetReply):
        if (self.isConnected() == False):
            self.log.write("Connection to " + self.host + " is lost or not yet established",'ERROR')
            raise Exception
        try:
            try :
                recv=''
                while True:
                    recvString=self.sock.recv(4096)
                    if  not recvString : break
                    recv+=recvString
            except socket.error as e:
                self.log.write("Socket Unable to read reply : "+ os.strerror(e.errno),'ERROR')
        except:
            self.log.write("Reply received", 'NOTIFICATION')

        try:
            packetReply.clearPacket()
            receivedBuffer = array('c')
            receivedBuffer.fromstring(''.join(recv))
            packetReply.packetRead(receivedBuffer)
            if self.isReplyPacketHeaderValid(packetReply):
                self.log.write("Reply header received is valid. Its type is: " + packetReply.getReplyType(), 'NOTIFICATION')
            else:
                raise StandardError("Invalid reply packet header")
        except StandardError as s:
            raise s
        except Exception:
            raise e
    

        

         
    def isReplyPacketHeaderValid(self,reply):
        try:
            if reply.isStartOfFrameCorrect() == False:
                self.log.write("Illegal start of frame !",'ERROR')
                return False
            elif reply.isMajorVersionSupported() == False:
                self.log.write("Major version not supported !",'ERROR')
                return False
            elif reply.isMinorVersionSupported() == False:
                self.log.write("Minor version not supported !",'ERROR')
                return False
            elif reply.isReplyPacket() == False :
                self.log.write("Is is not a reply packet !",'ERROR')
                return False
            return True
        except Exception:
            self.log.write("isReplyPacketHeaderValid : Catch exception:",'ERROR')
            raise Exception
    
        
    def handleProtocolError(self,replyPacket):
        try:
            if replyPacket.getPayload() == None:
                self.log.write("Empty Payload => unable to handle Protocol error ",'ERROR')
                raise Exception
 
            err_code=unpack_from('B',replyPacket.getPayload(),0)[0]

            if replyPacket.getNumBytesPayload() != 1: 
                self.log.write("RCI protocol error: illegal number of bytes for protocol error: one byte is needed",'ERROR')
                raise Exception
            elif err_code not in replyPacket.RCI_REPLY_PROT_ERROR.values():
                self.log.write("RCI protocol error: received Protocol error is not supported",'ERROR')
                raise Exception
            else:
                self.log.write("RCI protocol error \"" + replyPacket.getProtocolErrorType(err_code)+ "\"" ,'ERROR')
                raise Exception

        except Exception:
            self.log.write("handleProtocolError : Exception catched",'ERROR')
            raise Exception
    
    def sendReadRequest(self,request,reply):
        try:
            if request.getBaseTypeID() !=  request.RCI_REQ_BASE_ID['READ_REQUEST']:
                raise ValueError

            self.sendRequest(request)
            self.ReceiveReply(reply)
            if reply.getReplyId() == reply.RCI_REPLY_ID['RdRqReply']:
                self.log.write("Receiving reply for successful read request with data carried in payload",'NOTIFICATION')
            elif reply.getReplyId() == reply.RCI_REPLY_ID['ProtocolErr']:
                self.log.write("Receiving reply Protocol Error!!!",'ERROR')
                self.handleProtocolError(reply)
            else:
                self.log.write("RCI illegal reply id",'ERROR')
                raise Exception

        except Exception:
            self.log.write("sendReadRequest catch Exception",'ERROR')
            raise Exception
        except ValueError:
            self.log.write("sendReadRequest catch ValueError: Packet request is not not a read request")
            raise ValueError

    def sendWriteRequest(self,request,reply):
        try:
            if request.getBaseTypeID() !=  request.RCI_REQ_BASE_ID['WRITE_REQUEST']:
                raise ValueError

            self.sendRequest(request)
            self.ReceiveReply(reply)
            if reply.getReplyId() == reply.RCI_REPLY_ID['WrRqSuccess']:
                self.log.write("Receiving reply for successful write request",'NOTIFICATION')
            elif reply.getReplyId() == reply.RCI_REPLY_ID['WrRqFailed']:
                self.log.write("Receiving reply for failed write request",'WARNING')
                raise Exception
            elif reply.getReplyId() == reply.RCI_REPLY_ID['ProtocolErr']:
                self.log.write("Receiving reply Protocol Error!!!",'ERROR')
                self.handleProtocolError(reply)
            else:
                self.log.write("RCI illegal reply id",'ERROR')
                raise Exception
        except Exception:
            self.log.write("sendWriteRequest catch Exception",'ERROR')
            raise Exception
        except ValueError:
            self.log.write("sendWriteRequest catch ValueError: Packet request is not not a write request")
            raise ValueError

    def sendCmdExecutionRequest(self,request,reply):
        try:
            if request.getBaseTypeID() !=  request.RCI_REQ_BASE_ID['CMD_EXE_REQUEST']:
                raise ValueError

            request.setNumBytesPayload(0)
            self.sendRequest(request)
            self.ReceiveReply(reply)
            if reply.getReplyId() == reply.RCI_REPLY_ID['CmdRqSuccess']:
                self.log.write("Receiving reply for successful command execution request",'NOTIFICATION')
            elif reply.getReplyId() == reply.RCI_REPLY_ID['CmdRqFailed']:
                self.log.write("Receiving reply for failed command execution request",'WARNING')
                raise Exception
            elif reply.getReplyId() == reply.RCI_REPLY_ID['ProtocolErr']:
                self.handleProtocolError(reply)
            else:
                self.log.write("RCI illegal reply id",'ERROR')
                raise Exception
        except Exception:
            self.log.write("sendCmdExecutionRequest catch Exception",'ERROR')
            raise Exception
        except ValueError:
            self.log.write("sendCmdExecutionRequest catch ValueError : Packet request is not not a command execution request")
            raise ValueError

    def transmit(self,request,reply):
        try:
            if request.getBaseTypeID() == request.RCI_REQ_BASE_ID['READ_REQUEST']:
                self.sendReadRequest(request, reply)
            elif request.getBaseTypeID() ==  request.RCI_REQ_BASE_ID['WRITE_REQUEST']:
                self.sendWriteRequest(request, reply)
            elif request.getBaseTypeID() ==  request.RCI_REQ_BASE_ID['CMD_EXE_REQUEST']:
                self.sendCmdExecutionRequest(request, reply)
            else :
                self.sendReadRequest(request, reply)
                #raise StandardError("Illegal request base id")
        except StandardError:
            self.log.write("transmit : Catch StandardError",'ERROR')
            raise StandardError
        except Exception:
            self.log.write("transmit : catch Exception",'ERROR')
            raise Exception